/*

Subscribes to events sns topic and sends email notifications to appropriate users

*/
const env = process.env;
const fnName = env.AWS_LAMBDA_FUNCTION_NAME;
console.log(`Loading function ${fnName}...`);

const appId = 'com.endpointclosinginc.endpoint';
const settings = {
  gcm: {
    id: env.GOOGLE_API_KEY,
  },
  apn: {
    token: {
      key: env.APNS_CERT.replace(/\\n/g, '\n'),
      keyId: env.APNS_KEY_ID,
      teamId: '4L9GU5N75G',
    },
    production: true,
  },
};

const PushNotifications = require('node-pushnotifications');
const push = new PushNotifications(settings);

const firebaseAdmin = require('firebase-admin');

const firebaseConfig = {
  "type": env.TYPE,
  "project_id": env.PROJECT_ID,
  "private_key_id": env.PRIVATE_KEY_ID,
  "private_key": env.PRIVATE_KEY,
  "client_email": env.CLIENT_EMAIL,
  "client_id": env.CLIENT_ID,
  "auth_uri": env.AUTH_URI,
  "token_uri": env.TOKEN_URI,
  "auth_provider_x509_cert_url": env.AUTH_PROVIDER_X509_CERT_URL,
  "client_x509_cert_url": env.CLIENT_X509_CERT_URL,
}
const firebase = firebaseAdmin.initializeApp({
  credential: admin.credential.cert(firebaseConfig),
  databaseURL: env.DATABASE_URL
})

const isLocal = !!env.LOCALSTACK_HOSTNAME;
const isProduction = env.NODE_ENV === 'production';

const AWS = require('aws-sdk');
const config = new AWS.Config({
  accessKeyId: env.AWS_ACCESS_KEY_ID,
  secretAccessKey: env.AWS_SECRET_ACCESS_KEY,
  region: env.AWS_REGION,
});
const ddb = isLocal
  ? new AWS.DynamoDB({ endpoint: `http://${env.LOCALSTACK_HOSTNAME}:4569` })
  : new AWS.DynamoDB();

const { executer, helpers } = require('@endpoint/lambda-lib');
const h = helpers;
const he = executer;

exports.lambda_handler = async (event, context, callback) => {
  console.log(`event: ${h.obj2Str(event)}`);
  context.callbackWaitsForEmptyEventLoop = false;
  const snsData = he.validEventHandler(event);
  const messageAttributes = snsData.MessageAttributes;
  const principalTargets = JSON.parse(messageAttributes.principalTargets.Value);

  if (
    !he.targetsExistForExecuter(principalTargets, ['preferred_notifications'])
  ) {
    callback(
      null,
      `(${fnName}) No targets for this executer: ${JSON.stringify(
        principalTargets,
      )}`,
    );

    return `(${fnName}) No targets for this executer: ${JSON.stringify(
      principalTargets,
    )}`;
  }

  // Preferred not to go through getEventData as no roundtrip to db required
  const eventData = {
    transactionId: messageAttributes.transactionId.Value,
    principalId: messageAttributes.principalId.Value,
    copy: JSON.parse(messageAttributes.copy.Value),
    eventName: messageAttributes.eventName.Value,
  };

  // TODO abstract this out into helpersExecuter
  let res;

  try {
    const transactionRoles = await h.getTransactionRoles(
      ddb,
      eventData.transactionId,
    );

    let _preferredPushUserIds = [];

    Object.keys(transactionRoles).forEach(role => {
      // check if principal is a certain role, and proceed with that principalRole
      // hint: principal can play multiple roles :-)
      if (transactionRoles[role].includes(eventData.principalId)) {
        const principalRole = /Agent/.test(role) ? 'agent' : role.slice(0, -1); // role is pluralized in transactionRoles

        _preferredPushUserIds.push(
          ...he.parseTargetUserObjs(
            principalTargets[principalRole]['preferred_notifications'],
            transactionRoles,
            eventData.principalId,
            role,
          ),
        );
      }
    });

    const preferredPushUserIds = _preferredPushUserIds
      .map(obj => obj.userId)
      .unique();

    res = await pushHandler(preferredPushUserIds, eventData);
  } catch (err) {
    const message = err && typeof err === 'object' ? err.stack || err : err;

    console.log(`(${fnName}) failed: ${message}`);
    callback(null, `(${fnName}) failed: ${message}`);
  }

  console.log(`(${fnName}) succeeded: ${res}`);
  callback(null, `(${fnName}) succeeded: ${res}`);
};

/* Helpers localized to email executer */

function pushHandler(targetIds, eventData) {
  console.log(`pushHandler: targetIds: ${JSON.stringify(targetIds)}`);
  console.log(`pushHandler: eventData: ${JSON.stringify(eventData)}`);
  return Promise.all(
    targetIds.map(targetId => {
      return new Promise(async (resolve, reject) => {
        try {
          h.getUser(ddb, targetId)
          .then(async user => {
            if (shouldSendPush(user)) {
              const eventCopy =
                user.user_id === eventData.principalId
                  ? eventData.copy.principal
                  : eventData.copy.other;
              // using SMS copy
              console.log(
                `pushHandler: call pushTransport: user: ${JSON.stringify(
                  user,
                )}`,
              );
              console.log(
                `pushHandler: call pushTransport: eventCopy: ${JSON.stringify(
                  eventCopy,
                )}`,
              );
              await pushTransport(
                user.devices,
                eventData.eventName,
                eventCopy.sms,
              );

              resolve(`Resolving push-handler 🏁🏁🏁🏁`)
            } else {
              resolve(`Avoided sending push notifications to ${
                user.email
              } because:
                i. user does not have any device tokens stored
              `);
            }
          })
        }
        catch(err){
          console.log(`Error in push-executer index.pushHandler method ${err}`)
          throw err
        }
      });
    }),
  );
}

async function pushTransport(devices, pushTitle, pushMessage) {
  console.log('Sending push notification');
  console.log(`tokens: ${JSON.stringify(devices)}`);
  console.log(`pushTitle: ${JSON.stringify(pushTitle)}`);
  console.log(`pushMessage: ${JSON.stringify(pushMessage)}`);

  const iOSdevices = devices.filter(i => i.platform.toLowerCase() === 'ios' && i.token.length > 0);
  const androidDevices = devices.filter(i => i.platform.toLowerCase() === 'android' && i.token.length > 0);

  console.log(`iOS Tokens: ${JSON.stringify(iOSdevices)}`)
  console.log(`androidDevices Tokens: ${JSON.stringify(androidDevices)}`)

  const iOSTokenArray = iOSdevices.length ? iOSdevices.map(t => t.token) : [];
  const androidTokenArray = androidDevices.length ? androidDevices.map(t => t.token) : [];

  const iOSPayload = { title : pushTitle , body : pushMessage }
  const androidPayload = { notification : { title: pushTitle , body : pushMessage }}

  await iOSHandler(iOSTokenArray, iOSPayload);  
  await androidHandler(androidTokenArray, androidPayload);
 
  return
}

function shouldSendPush(user) {
  const validDevices = user.devices ? user.devices.filter(i => i.token && i.platform && i.platform.length && i.token.length) : [];
  const shouldSend = validDevices.length ? true : false;
  return shouldSend;
}


async function iOSHandler(tokens, payload){
  if(tokens.length > 0) {
    const res = await push.send(tokens, {    
      topic: appId, //golab var
      ...payload,
      })
      console.log(`iOSHandler`)
  } else {
    console.log(`iOSHandler - No iOS device tokens provided`)
  }
  return
}

async function androidHandler(tokens, payload){
  if(tokens.length > 0 ){
    const res = await firebase.messaging().sendToDevice(tokens, payload , { dryRun : false})

    if(res.failureCount > 0 ){
      console.log(`☠️ androidHandler - ${res.failureCount} Notifications failed to sent ...`)
      res.results.forEach(r => {
        if(r.error){
          console.log(`Error: ${r.error}`)
        }
      })
    }
    if(res.successCount > 0 ){
      console.log(`✅ androidHandler - ${res.successCount} Notifications succesfully sent ...`)
    }
    console.log(`Exiting Android Handler multicastId: ${res.multicastId}`)
  }
  else {
    console.log(`AndroidHandler - No android device tokens provided`)
  }
  return
}